/* ccgo: go/go.hh
 *
 * Copyright 2004 Chun-Chung Chen <cjj@u.washington.edu>
 *
 * This program is free software; you can redistribute it and'or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#ifndef GO_GO_HH
#define GO_GO_HH

#include <vector>
#include <map>
#include <set>
#include <string>

namespace ccgo
{
	class Error
	{
	public:
		Error(std::string message);
		const std::string & get_message();
	private:
		std::string s;
	};

	// implement: simple square board
	class Board
	{
	public:
		enum State {
			STATE_EMPTY = 0,
			STATE_BLACK = 1,
			STATE_WHITE = 2
		};
		enum Terri {
			TERRI_NEUTRAL = 0,
			TERRI_BLACK = 1,
			TERRI_WHITE = 2
		};
		enum {
			NO_LOC = - 1
		};
		Board();
		Board(int size);
		Board(const Board & board);
		~Board();
		void copy(const Board & board);
		virtual void clear(); // clear board
		int get_size() const;
		State get_state(int loc) const; // get State at "loc"
		virtual void set_state(int loc, State state); // set State at "loc" to "state"
		std::vector<int> neighbors(int loc) const; // return neighboring sites of "loc"
		int loc_range() const; // return valid range of Loc
		int flip_axis(int x) const;
		int xy_to_loc(int x, int y) const;
		void loc_to_xy(int loc, int & x, int & y) const;

		// housekeeping functions

		void get_kill(int loc, std::vector<int> & kill, State put_state) const;
		// put sites with killed stones when place "put_state" at "loc" in "kill"
		int qi_group(int loc, std::vector<int> & group, State state, bool * marking_field) const;
		// return number of qi for the "group" including site at "loc" with "state", skip
		// sites which has been marked in "marking_field"

		void score_terri(Terri * terri) const;
		void find_terri(int loc, std::vector<int> & group, bool & w_seen, bool & b_seen, Terri * terri, bool * marks) const;
		void hang_group(int loc, State state, std::vector<int> & qi, Terri * terri, bool * marks) const;
		void get_group(int loc, std::vector<int> & group) const;
		void dead_group(int loc, std::vector<int> & group) const;
		void terri_group(int loc, State state, std::vector<int> & group, bool * marks) const;

		// functions provides a compact data structure of current board configuration
		int compact_size() const;
		void compact_get(void * data) const;

		bool loc_is_star(int loc) const;
		virtual void set_handi(int handi);
		std::string loc_to_str(int loc) const;
	private:
		int z;     // size
		State * b; // board
	};

	enum Move {
		MOVE_PASS = - 1,
		MOVE_NONE = - 100
	};

	class Position :
		public Board
	{
	public:
		Position();
		Position(int size);
		Position(const Position & position);
		void copy(const Position & position);
		virtual void clear();
		enum Turn {
			TURN_BLACK = 1,
			TURN_WHITE = 2
		};
		Turn get_turn() const;
		virtual void set_turn(Turn turn);
		const std::set<int> & get_ko() const;
		virtual void set_ko(const std::set<int> & ko);

		bool loc_is_ko(int loc) const;
		virtual void set_handi(int handi);
		virtual void flip_turn();
		static Turn flip_turn(Turn turn);
	private:
		Turn t; // turn
		std::set<int> k; // ko list
	};

	class Delta
	{
	public:
		Delta();
		Delta(const Delta & delta);
		void apply(Position & position) const;
		void revert(Position & position) const;
		void add_change(int loc, Board::State old_state, Board::State new_state);
		void del_change(int loc);
		void set_turn_flip(bool turn_flip);
		void set_ko_old(std::set<int> old_ko);
		void set_ko_new(std::set<int> new_ko);

		bool get_turn_flip() const;
		bool is_null();
		bool is_changed(int loc);
		Board::State get_new_state(int loc);

		int num_change();
	private:
		struct D {
			int loc;
			Board::State os; // old state
			Board::State ns; // new state
		};
		std::vector<D> dl; // change list
		bool tf; // turn flip?
		std::set<int> ok;
		std::set<int> nk;
	};

	class Prop
	{
	public:
		static int the_id();
		virtual int get_id();
		virtual Prop * dup() const;
	protected:
		class Id // class counter
		{
		public:
			Id();
			int operator()();
		private:
			static int c;
			int v;
		};
	private:
		static Id id;
	};

	class Comment :
		public Prop
	{
	public:
		static int the_id();
		virtual int get_id();
		virtual Prop * dup() const;
		
		const std::string & get_comment();
		void set_comment(const std::string & text);
	private:
		std::string c;
		static Id id;
	};

	enum Mark {
		MARK_NONE = 0,
		MARK_CIRCLE,
		MARK_TRIANGLE,
		MARK_SQUARE,
		MARK_CROSS,
		MARK_SELECT,
		MARK_W_TERRITORY,
		MARK_B_TERRITORY
	};

	class Marking :
		public Prop
	{
	public:
		static int the_id();
		virtual int get_id();
		virtual Prop * dup() const;
		void clear_marks();
		void add_mark(int loc, Mark mark);
		Mark get_mark(int loc);
		const std::map<int, Mark> & get_marks() const;
	private:
		std::map<int, Mark> m;
		static Id id;
	};

	class Labels :
		public Prop
	{
	public:
		static int the_id();
		virtual int get_id();
		virtual Prop * dup() const;
		void clear_labels();
		void add_label(int loc, const std::string & label);
		const std::string & get_label(int loc);
		const std::map<int, std::string> & get_labels() const;
	private:
		std::map<int, std::string> lb;
		static Id id;
	};

	class Game;

	class Node
	{
	public:
		Node(Node * parent);
		virtual ~Node();
		void copy(const Node & node);
		Node * up();
		int down_num(); // number of down links
		Node * down(int index);
		Game * game(); // game node
		bool set_move(int move, const Delta & delta);
		void clear_move();
		int get_move() const;
		const Delta * get_delta() const;
		Delta * get_delta();
		int add_child();
		int get_index();
		void del_child(Node * node);
		void del_child(int index);
		int get_num();
		void set_num(int num);
		Prop * get_prop(int id);
		void set_prop(int id, Prop * prop);
	private:
		Node * u; // up link
		std::vector<Node *> d; // down links
		int m; // Move
		Delta * c; // Changes by the move
		Node * g; // Game
		int n; // Move number
		std::map<int, Prop *> p;
	};

	class Walk :
		public Position
	{
	public:
		Walk();
		~Walk();
		virtual void set_node(Node * node); // attach the walk to a node
		Node * where() const;
		void up();
		void down(int index);
		int get_level();

		// virtual void clear();

		bool game_can_init();
		int game_add_move(int move);
		int game_add_setup();
		void game_set_turn(Turn turn);
		void game_set_site(int loc, State state);
		void game_set_handi(int handi);
		void game_add_handi(int loc);

		int get_b_cap();
		int get_w_cap();
	protected:
		virtual void init_game(Game * game);
		virtual void apply_node(Node * node);
		virtual void revert_node(Node * node);
	private:
		Node * n;
		int l; // level
		int cap_b;
		int cap_w;
	};

	struct Rule
	{
		Rule();
		bool opt_overwrite_stone;
		bool opt_ignore_ko;
		bool opt_allow_suicide;
		enum Ko {
			KO_SIMPLE = 0,
			KO_NONE = 1,
			KO_POSITIONAL = 2,
			KO_SITUATIONAL = 3
		};
		Ko opt_ko_style;
	};

	class Game :
		public Node
	{
	public:
		Game();
		~Game();

		void copy(const Game & game);
		
		void set_init(const Position & position);
		const Position * get_init() const;
		Position * get_init();
		bool check(const Walk & walk, int move);
		Delta make_delta(const Walk & walk, int move);
		const Rule & get_rule() const;
		void set_rule(const Rule & rule);
		void add_walk(Walk * walk);
		void del_walk(Walk * walk);

		void set_name(std::string name);
		const std::string & get_name() const;
		void set_komi(double komi);
		double get_komi() const;
	private:
		Position * p; // initial position
		Rule r;
		bool h; // handicap
		std::vector<Walk *> w;

		// Game parameters
		std::string nm;
		double km;
	};

}
#endif // GO_GO_HH
